;-----------------------------------------------------------------
;
;	AtariKlang standalone EXE player
;	Leonard/Oxygene, May 2022
;
;-----------------------------------------------------------------

; amigahub atari: 50s
; amigahub amiga: 95s

PRG_HEADER				equ	1
STE_SOUND_CHECK			equ	1
EQUALIZER				equ	1
SOFT_SYNTH				equ	1

RAW_SAMPLES				equ	(AK_EXT_SMP_LEN>0)
PROGRESS_BAR			equ	(AK_FINE_PROGRESS>0)

BSS_ESTIMATED_SIZE		equ	(64*1024+AK_SMP_LEN+32*1024)

;------------------------------------------------------------
	
		IFNE	PRG_HEADER
			dc.w	$601a
			dc.l	mEnd-mStart		; text seg
			dc.l	0				; data seg
			dc.l	BSS_ESTIMATED_SIZE		; bss 
			dc.l	0				; symbol table len
			dc.l	0				; reserved
			dc.l	0				; flags
			dc.w	-1				; non zero means "absolute" (no relocation table!)
		ENDC
	
mStart:		
			lea		mStart(pc),a0
			add.l	#bssZone-mStart,a0
			lea		mRelocTable(pc),a1
.rLoop:		move.l	(a1),d0					; chunk size
			bmi.s	.over
			move.l	a0,(a1)+
			add.l	d0,a0
			bra.s	.rLoop
.over:
			move.l	pUserStack(pc),a7

			clr.l	-(a7)
			move.w	#32,-(a7)
			trap	#1

		IFNE	STE_SOUND_CHECK
			move.l	$5a0.w,d0
			beq.s	.noSTE
			move.l	d0,a0
.cloop:		move.l	(a0)+,d0
			beq.s	.noSTE
			cmp.l	#'_SND',d0
			beq.s	.found
			addq.w	#4,a0
			bra.s	.cloop
.found:		btst.b	#1,3(a0)
			bne.s	.valid
.noSTE:		clr.w	-(a7)
			trap	#1
.valid:
		ENDC

			clr.w	-(a7)
			moveq	#-1,d0
			move.l	d0,-(a7)
			move.l	d0,-(a7)
			move.w	#5,-(a7)
			trap	#1

			bra		jmpOverAk2asm


	IFNE	SOFT_SYNTH
ak2asm:		include	"exemusic.asm"
			even
	ENDC	

jmpOverAk2asm:

			move.l	pStack(pc),a7
			bsr		clearScreen

			clr.w	$ffff8240.w

	IFNE	PROGRESS_BAR
			move.w	#$333,$ffff8242.w
			move.w	#$777,$ffff8246.w
			move.w	#320,d0
			moveq	#0,d1				; screen offset
			bsr		drawProgressBar
	ELSE
			move.w	#$400,$ffff8240.w
	ENDC

			move.w	#$2700,sr
			clr.l	$fffffa06.w
			clr.l	$fffffa12.w
	IFNE	EQUALIZER
			bset	#0,$fffffa07.w
			bset	#0,$fffffa13.w
			sf		$fffffa1b.w
	ENDC
			bclr	#3,$fffffa17.w
			lea		vblIrq(pc),a0
			move.l	a0,$70.w
			move.w	#$2300,sr

		; init audio replayer
			bsr		paulaInstall

		IFNE	SOFT_SYNTH
			move.l	pSoftSynthBank(pc),a0
			addq.w	#4,a0
			lea		progressBar(pc),a1
			st		(a1)
			move.l	pTmpBuffer(pc),a1
			lea		mStart(pc),a2
			add.l	#rawSamples-mStart,a2
			lea		progressBarCounter(pc),a3
			bsr		AK_Generate
			lea		progressBar(pc),a1
			sf		(a1)
			clr.w	$ffff8240.w
		ENDC
			bsr		clearScreen
			
			lea		lspMusic(pc),a0
		IFNE	SOFT_SYNTH
			move.l	pSoftSynthBank(pc),a1
		ELSE
			movea.l	a0,a1
			add.l	#lspBank-lspMusic,a1
		ENDC
			bsr		LSP_MusicDriver
			lea		pCurrentBPM(pc),a1
			move.l	a0,(a1)

		IFNE	EQUALIZER
			lea		colorsTab(pc),a0
			moveq	#16*4-1,d0
.amigaConv:	move.w	(a0),d1
			move.w	d1,d2
			andi.w	#$eee,d1
			andi.w	#$111,d2
			lsr.w	#1,d1
			lsl.w	#3,d2
			or.w	d2,d1
			move.w	d1,(a0)+
			dbf		d0,.amigaConv
			
			lea		vblEqualizer(pc),a0
			move.l	a0,$70.w
		ENDC

			
		move.w	#$fff,$ffff8242.w

mainLoop:	
			move.w	$468.w,d0
.vsync:		cmp.w	$468.w,d0
			beq.s	.vsync

			lea		noteOnMask(pc),a0
			clr.w	(a0)
			move.l	pCurrentBPM(pc),a0
			moveq	#0,d0
			move.w	(a0),d0				; music driver rate (in BPM)
			add.w	d0,d0
			divu.w	#5,d0				; music driver rate (in Hz)
			bsr		paulaTick


		IFNE	EQUALIZER
			
			move.w	noteOnMask(pc),d0
			lea		vuMeters(pc),a1
			lea		colorsTab(pc),a2
			lea		copperList(pc),a4
			moveq	#4-1,d7
.vLoop:		btst	d7,d0
			beq.s	.noNote
			move.w	#16*2,(a1)
.noNote:	tst.w	(a1)
			beq.s	.low
			subq.w	#2,(a1)
.low:		movea.l	a2,a3
			add.w	(a1)+,a3
			move.w	(a3),(a4)
			lea		16*2(a2),a2
			lea		4*2(a4),a4
			dbf		d7,.vLoop
			
		ENDC
			
			
			bra		mainLoop
		

progressBarCounter:
			dc.l	0
pCurrentBPM:
			dc.l	0

clearScreen:
			move.l	$44e.w,a0
			move.w	#200*160/16-1,d0
			moveq	#0,d1
.cls:		move.l	d1,(a0)+
			move.l	d1,(a0)+
			move.l	d1,(a0)+
			move.l	d1,(a0)+
			dbf		d0,.cls
			rts


	IFNE	PROGRESS_BAR
; d0: bar size
; d1: screen offset
drawProgressBar:
			move.l	$44e.w,a1
			lea		99*160(a1),a1
			add.w	d1,a1
			move.w	d0,d1
			lsr.w	#4,d1
			beq.s	.nofull
			subq.w	#1,d1
.floop:		move.w	#-1,(a1)
			move.w	#-1,160(a1)
			addq.w	#8,a1
			subi.w	#16,d0
			dbf		d1,.floop
.nofull:	tst.w	d0
			beq.s	.skip
			move.l	#$ffff<<15,d1
			lsr.l	d0,d1
			move.w	d1,(a1)
			move.w	d1,160(a1)
.skip:
			rts
	ENDC
			
			
vblIrq:
	IFNE	PROGRESS_BAR
			movem.l	d0-a6,-(a7)			; on purpose NOT save a0
			lea		progressBar(pc),a2
			tst.b	(a2)
			beq.s	.over
			
			move.l	progressBarCounter(pc),d0
			lsr.l	#8,d0
			mulu.w	#319,d0
			divu.w	#AK_FINE_PROGRESS_LEN>>8,d0
			
			moveq	#2,d1
			bsr		drawProgressBar
			
.over:

			movem.l	(a7)+,d0-a6
	ENDC
			addq.w	#1,$468.w
			rte

	IFNE	EQUALIZER
vblEqualizer:

			movem.l	d0-a6,-(a7)
			
			lea		timerB1(pc),a1
			move.l	a1,$120.w
			move.b	#52,$fffffa21.w
			move.b	#8,$fffffa1b.w
			lea		copperList(pc),a1
			move.w	(a1)+,$200.w
			move.l	a1,$202.w
			

			movem.l	(a7)+,d0-a6

			addq.w	#1,$468.w
			rte
			
timerB1:	move.w	$200.w,$ffff8240.w
			bmi.s	.over
			pea		(a1)
			move.l	$202.w,a1
			move.b	#0,$fffffa1b.w
			move.w	(a1)+,$fffffa20.w
			move.b	#8,$fffffa1b.w
			move.w	(a1)+,$200.w
			move.l	a1,$202.w
			move.l	(a7)+,a1
			rte
.over:		move.b	#0,$fffffa1b.w
			rte
			
copperList:
			dc.w	$000		; color
			dc.w	16			; lines to wait
			dc.w	$000		; color
			dc.w	8

			dc.w	$000		; color
			dc.w	16			; lines to wait
			dc.w	$000		; color
			dc.w	8

			dc.w	$000		; color
			dc.w	16			; lines to wait
			dc.w	$000		; color
			dc.w	8

			dc.w	$000		; color
			dc.w	16			; lines to wait

			dc.w	$8000

colorsTab:	dc.w	$000,$000, $100, $200, $211, $311, $322, $422, $533, $644, $755, $866, $977, $b99, $dbb, $fdd
			dc.w	$000,$000, $110, $220, $221, $331, $332, $442, $553, $664, $775, $886, $997, $bb9, $ddb, $ffd
			dc.w	$000,$000, $010, $020, $121, $131, $232, $242, $353, $464, $575, $686, $797, $9b9, $bdb, $dfd
			dc.w	$000,$000, $001, $002, $112, $113, $223, $224, $335, $446, $557, $668, $779, $99b, $bbd, $ddf

vuMeters:	dc.w	0,0,0,0

	ENDC
	
progressBar:
			dc.w	0
lastValue:	dc.w	0
noteOnMask:	dc.w	0

			
			include "LightSpeedPlayerAtari.asm"
			even

			include "PaulaEmulatorTiny.asm"
			even


	; easy relocatable pointers
mRelocTable:
pStackBuffer:		dc.l	1024									; 1 kb
pStack:				dc.l	0
pUserStackBuffer:	dc.l	1024									; 1 kb
pUserStack:			dc.l	0
pVolumeTableBuffer:	dc.l	65*256+256								; 17 kb
pFrequenceTable:	dc.l	PERMAX*4								; 14 kib
pSampleBuffer:		dc.l	PAULA_MIXING_BUFFER_SIZE*NB_BLOCK		; 4 kib
	IFNE	SOFT_SYNTH
pSoftSynthBank:		dc.l	AK_SMP_LEN								; colombia = 160 kb
pTmpBuffer:			dc.l	32*1024									; 32 kib
	ENDC
					dc.l	-1

lspMusic:			incbin	"AtariKlang.lsmusic"
					even
					
	IFEQ	SOFT_SYNTH
lspBank:			incbin	"AtariKlang.lsbank"
					even
	ELSE
rawSamples:			
		IFNE	RAW_SAMPLES
					incbin	"Isamp.raw"
					even
		ENDC
	ENDC
	
mEnd:

bssZone:
			
