;-----------------------------------------------------------------------------------
;
;	Emulation chip AMIGA Paula.
;
;  	Atari STE - 25Khz
;	Written by Leonard (Arnaud Carre) Dec 1996.
;
;
;	new 2006 version, cleaner and faster.
;	July 2007: relocatable code and CT60 compatible (MOVEP removed)
;
; Nov 2021 :) keep only STE code path for 4KiB demo 4k-MOD (Sillyventure 2021)
; May 2022 :) Makes it vasm compatible for AtariKlang
; 
;-----------------------------------------------------------------------------------

ASSERT							equ	0
MICROWIRE_SET					equ	0

PLAYERRATE						equ	50
PAULA_25K_NBSAMPLE_PER_BLOCK	equ	512			; way more than needed (25033/50), so we can skip frame sometimes and perfectly keep in sync without using interrupts
NB_BLOCK						equ	4
PAULA_MIXING_BUFFER_SIZE		equ	PAULA_25K_NBSAMPLE_PER_BLOCK*2
MIXER_CHUNK_SIZE				equ	8

paulaInstall:	

		move.b	#(1<<1),$ffff8921.w	; stereo, 25Khz	
		
		bsr	volumeBuild
		bsr	frequenceBuild

		move.w	#(PAULA_MIXING_BUFFER_SIZE*NB_BLOCK)/4-1,d0
		move.l	pSampleBuffer(pc),a0
.fill:	clr.l	(a0)+
		dbf		d0,.fill
		
	IFNE	MICROWIRE_SET
	;--------------------------------------
	; setup MicroWire
	;--------------------------------------
		lea		MicroWireData(pc),a0
		move.w	#$07ff,$ffff8924.w
		moveq	#3-1,d0
.wait1:	tst.w	$ffff8922.w
		bne.s	.wait1
		move.w	(a0)+,$ffff8922.w
		dbf		d0,.wait1
	ENDC
		move.l	pSampleBuffer(pc),a0
		lea		$ffff8900.w,a1

		move.l	a0,d0
		movep.l	d0,$1(a1)			; set start ad

		lea		PAULA_MIXING_BUFFER_SIZE*NB_BLOCK(a0),a0
		move.l	a0,d0
		movep.l	d0,$d(a1)			; set end address

		move.b	#%11,1(a1)			; start (bit0) and loop mode (bit1)
.over:

		lea		driverPhase(pc),a0
		clr.w	(a0)

		lea		.backToSuper(pc),a0
		move.l	a0,$80.w

		rts

.backToSuper:
		ori.w	#$2000,(a7)
		rte

	IFNE	MICROWIRE_SET
MicroWireData:
		dc.w	%0000010011000000+40	; master volume
		dc.w	%0000010101000000+20	; left volume
		dc.w	%0000010100000000+20	; right volume
	ENDC

volumeBuild:	
		move.l	pVolumeTableBuffer(pc),a0
		lea		255(a0),a0
		move.l	a0,d0
		clr.b	d0
		lea		pVolumeTable(pc),a0
		move.l	d0,(a0)
		move.l	d0,a0
		moveq	#0,d0
		
		moveq	#6+1,d3				; 6bit shift for volume, keep only 7bits per sample
		
.loopv:	moveq	#0,d1
.loops:	move.b	d1,d2
		ext.w	d2
		muls.w	d0,d2
		asr.w	d3,d2		; /64: Amiga volume max.
		move.b	d2,(a0)+
		addq.b	#1,d1
		bne.s	.loops
		addq.w	#1,d0
		cmpi.w	#65,d0
		bne.s	.loopv
		rts

PERMIN		equ	$71
PERMAX		equ	$d60


frequenceBuild:	
		move.l	pFrequenceTable(pc),a0
		moveq	#PERMIN-1,d0
.clear:	clr.l	(a0)+
		dbf	d0,.clear
		move.w	#PERMIN,d0
		move.w	#PERMAX-PERMIN-1,d1
		move.l	#9371195,d2	; (3579546(paula) * 65536(prec)) / 25033(MIXERFRQ)
		move.l	d2,d5
		clr.w	d5
		swap	d5
.compute:
		move.l	d5,d3
		divu	d0,d3
	IFNE	ASSERT
		cmpi.w	#2,d3
		bge		assertFunc
	ENDC
		move.w	d3,(a0)+	; high part
		move.w	d2,d3
		divu	d0,d3
		move.w	d3,(a0)+
.backClamp:
		addq.w	#1,d0
		dbf		d1,.compute
		rts

; input: d0 voice (0,1,2 or 3)
voiceMixProcess:
		pea		(a3)
		
		move.w	dmacon(pc),d3
		btst	d0,d3
		beq		.noDMA

		lsl.w	#4,d0
		lea		channela(pc),a0
		add.w	d0,a0
		lea		-16*4(a0),a1

		pea	(a1)

		move.w	8(a0),d3		; volume
		lsl.w	#8,d3

		move.l	pVolumeTable(pc),a2
		add.w	d3,a2				; Adresse volume table.
		move.l	a2,d2

		move.l	pFrequenceTable(pc),a2
		move.w	6(a0),d3
		
		cmpi.w	#PERMAX,d3
		bls.s	.clipFreq
		move.w #PERMAX-1,d3
.clipFreq:
		add.w	d3,d3
		add.w	d3,d3
		add.w	d3,a2

		; pointeurs de loop
		move.l	(a0),d3			; prochain repeat sample
		moveq	#0,d1
		move.w	4(a0),d1		; rep len en words

		add.l	d1,d1
		move.l	d3,d4
		add.l	d1,d4			; prochain end loop

		move.l	(a1),a0			; current sample pointer
		move.w	8(a1),d1		; current fixed int
		move.w	10(a1),d6		; last audat
		move.l	4(a1),a1		; current sample loop end

		exg		a3,a7

		MACRO	SAMPLE_MIX
			add.b	d6,(a7)+
			add.w	d0,d1
			bcc.s	.fcode\@
			move.b	(a0)+,d2
			cmpa.l	a0,a1
			bgt.s	.noLoop\@
			move.l	d3,a0
			move.l	d4,a1
.noLoop\@:	move.l	d2,a2
			move.b	(a2),d6
.fcode\@:
		ENDM

		MACRO	SAMPLE_MIX_HIGHFREQ
			add.b	d6,(a7)+
			move.b	(a0)+,d2
			move.l	d2,a2
			move.b	(a2),d6
			add.w	d0,d1
			bcc.s	.skip\@
			addq.w	#1,a0
.skip\@:	cmpa.l	a0,a1
			bgt.s	.noLoop\@
			move.l	d3,a0
			move.l	d4,a1
.noLoop\@:	
		ENDM


		moveq	#(PAULA_25K_NBSAMPLE_PER_BLOCK/MIXER_CHUNK_SIZE)/4-1,d7	; /4 to handle variable music driver rate (200hz)
		
		tst.w	(a2)+			; in this tiny version, only support high part at 0
		bne		.highFreq
		move.w	(a2),d0			; fractional part

.mixLoop
		REPT	MIXER_CHUNK_SIZE
			SAMPLE_MIX
		ENDR
		dbf		d7,.mixLoop
		bra		.endMix

.highFreq:

		move.w	(a2),d0			; fractional part
.mixLoop2
		REPT	MIXER_CHUNK_SIZE
			SAMPLE_MIX_HIGHFREQ
		ENDR
		dbf		d7,.mixLoop2

.endMix:


		exg		a3,a7

		; no need to test loop
		move.l	(a7)+,a2
		move.l	a0,(a2)+
		move.l	a1,(a2)+
		move.w	d1,(a2)+
		move.w	d6,(a2)+


.noDMA:
		move.l	(a7)+,a3
		rts

	IFNE	ASSERT
assertFunc:
		move.l	#$12345678,d7
		illegal
	ENDC


paulaTick:
		lea		driverRate(pc),a0
		move.w	d0,(a0)
		
	;---------------------------------------------------------------
	; check si on a un block libre complet
	;---------------------------------------------------------------
		lea		$ffff8900.w,a0
		move.b	$9(a0),d0
		swap	d0
		move.b	$b(a0),d0	
		lsl.w	#8,d0
		move.b	$d(a0),d0
		andi.l	#$003ffffe,d0
		move.l	pSampleBuffer(pc),a0
		sub.l	a0,d0					; offset dans buffer

		sub.w	renderingBlock(pc),d0	; - offset de rendering
		bmi.s	.okr					; si negatif alors rendering devant playing, c'est ok

		; si positif on doit checker si y a de la place
		cmp.w	#PAULA_MIXING_BUFFER_SIZE,d0
		blt		.skipMixing				; si on est trop pres du player on skip le mixing pour cette frame
		
.okr:


	;---------------------------------------------------------------
	; Mixage des 4 voies dans samp1
	;---------------------------------------------------------------

;-----------------------------------------------------------
.steMixing:
;-----------------------------------------------------------

		move.l	pSampleBuffer(pc),a3
		lea		PAULA_MIXING_BUFFER_SIZE(a3),a3
		add.w	renderingBlock(pc),a3

		moveq	#(PAULA_MIXING_BUFFER_SIZE/8)-1,d0
		moveq	#0,d1
.clrb:	move.l	d1,-(a3)
		move.l	d1,-(a3)
		dbf		d0,.clrb


		; a7 will be used for mixing, so set user mode
		move.w	#$300,sr


		move.w	#4,-(a7)
.simul200hz:

	;---------------------------------------------------------------
	; call music driver with variable BPM ( emulated at 200hz )
	;---------------------------------------------------------------
		lea		driverPhase(pc),a0
		move.w	driverRate(pc),d0
		add.w	d0,(a0)
		cmpi.w	#200,(a0)
		blt.s	.skipMusicDriver
		subi.w	#200,(a0)

		pea		(a3)
		lea		channela(pc),a6
		bsr		LSP_MusicDriver+4
		move.l	(a7)+,a3

.skipMusicDriver:

		moveq	#0,d0
		bsr	voiceMixProcess
		moveq	#3,d0
		bsr	voiceMixProcess
		addq.l	#1,a3				; 1 byte offset for left speaker
		moveq	#1,d0
		bsr	voiceMixProcess
		moveq	#2,d0
		bsr	voiceMixProcess
	
		subq.l	#1,a3
		lea		(PAULA_MIXING_BUFFER_SIZE/4)(a3),a3
		subq.w	#1,(a7)
		bne.s	.simul200hz
		addq.w	#2,a7

		; back to supervisor
		trap	#0

	;---------------------------------------------------------------
	; Avance les pointeurs pour prochain appel.
	;---------------------------------------------------------------
		lea		renderingBlock(pc),a0
		addi.w	#PAULA_MIXING_BUFFER_SIZE,(a0)
		andi.w	#PAULA_MIXING_BUFFER_SIZE*4-1,(a0)	; now PAULA_25K_NBSAMPLE_PER_BLOCK is power of 2!

.skipMixing:
		rts
		

; d6.w : dmacon
writeDmacon:
		movem.l	d0-d2/d6/a0-a2,-(a7)
		lea		dmacon(pc),a0
		btst	#15,d6
		bne.s	.set

	; Ici on clear des dma de sample: coupe chaque voie.
		not.w	d6
		and.w	d6,(a0)
		bra.s	.ok

	; Ici on allume des DMA de sample: restart chaque voie.
.set:		
		move.w	(a0),d0				; old DMACon
		or.w	d6,(a0)
		lea		channeld(pc),a0
		moveq	#4-1,d1

.loop:	btst	d1,d6
		beq.s	.skip
		btst	d1,d0				; if previous DMA was already set, do nothing
		bne.s	.skip

		move.l	(a0),a2
		move.l	a2,-16*4(a0)		; set sample adress.
		moveq	#0,d2
		move.w	4(a0),d2			; set sample len
		add.l	d2,d2
		add.l	d2,a2
		move.l	a2,4-16*4(a0)
		move.w	#0,8-16*4(a0)		; fixed part = 0

.skip:	lea	-16(a0),a0
		dbf	d1,.loop

.ok:	movem.l	(a7)+,d0-d2/d6/a0-a2
		rts

;-----------------------------------------------------------------					
renderingBlock:		dc.w	0
pVolumeTable:		dc.l	0
dmacon:				dc.w	0		; DMACON ($dff096)
driverPhase:		dc.w	0
driverRate:			dc.w	0

	; Custom chip AMIGA:
	;	$dff0a0.L : Sample start adress
	;	$dff0a4.W : Sample length (in words)
	;	$dff0a6.W : Channel period.
	;	$dff0a8.W : Channel Volume.

pSampA		ds.l	1   ; current sample ptr
			ds.l	1	; end sample ptr
			ds.w	1	; fixed point
			ds.w	1	; audxdat
			ds.l	1	; pad

pSampB		ds.l	1   ; current sample ptr
			ds.l	1	; end sample ptr
			ds.w	1	; fixed point
			ds.w	1	; audxdat
			ds.l	1	; pad

pSampC		ds.l	1   ; current sample ptr
			ds.l	1	; end sample ptr
			ds.w	1	; fixed point
			ds.w	1	; audxdat
			ds.l	1	; pad

pSampD		ds.l	1   ; current sample ptr
			ds.l	1	; end sample ptr
			ds.w	1	; fixed point
			ds.w	1	; audxdat
			ds.l	1	; pad


channela:	; $dff0a0
		dc.l	0
		dc.w	0
		dc.w	0
		ds.b	8		; Custom chip canal 0
channelb:
		dc.l	0
		dc.w	0
		dc.w	0
		ds.b	8		; Custom chip canal 0
channelc:
		dc.l	0
		dc.w	0
		dc.w	0
		ds.b	8		; Custom chip canal 0
channeld:
		dc.l	0
		dc.w	0
		dc.w	0
		ds.b	8		; Custom chip canal 0
		
				even
