#include "support/Iswitch.h"
#ifdef executable

#include "support/gcc8_c_support.h"
#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/graphics.h>
#include <graphics/gfxbase.h>
#include <graphics/view.h>
#include <exec/execbase.h>
#include <graphics/gfxmacros.h>
#include <hardware/custom.h>
#include <hardware/dmabits.h>
#include <hardware/intbits.h>

#include "Iset.h"

#define BACKGROUND_COLOR 0x000

// Number of instruments to precalc before starting playing
// Default is 32. Use this for keeping to a 30 second precalc
// limit in compos and putting instruments used in initial
// patterns at lower instrument numbers ;)
#define EARLYPLAY_INSTRUMENTS_COMPLETE 32

// Globals

struct ExecBase *SysBase;
volatile struct Custom *custom;
struct DosLibrary *DOSBase;
struct GfxBase *GfxBase;

// system backups
static UWORD SystemInts;
static UWORD SystemDMA;
static UWORD SystemADKCON;
volatile static APTR VBR=0;
static APTR SystemIrq;
struct View *ActiView;

volatile UWORD InstrumentsCurrent = 0;
volatile UWORD ForceQuit = 0;

static void * ModAdr = NULL;
static void * ImpAdr = NULL;

static short decayTable[128] = { 32767, 32767, 32767, 16384, 10922, 8192, 6553, 4681, 3640, 2978, 2520, 2048, 1724, 1489, 1310, 1129, 992,
885, 799, 712, 642, 585, 537, 489, 448, 414, 385, 356, 330, 309, 289, 270, 254, 239, 225, 212, 201, 190, 181, 171, 163, 155, 148, 141, 134,
129, 123, 118, 113, 108, 104, 100, 96, 93, 89, 86, 83, 80, 77, 75, 72, 70, 68, 65, 63, 61, 60, 58, 56, 54, 53, 51, 50, 49, 47, 46, 45, 44,
43, 41, 40, 39, 38, 38, 37, 36, 35, 34, 33, 33, 32, 31, 30, 30, 29, 29, 28, 27, 27, 26, 26, 25, 25, 24, 24, 23, 23, 22, 22, 22, 21, 21, 20,
20, 20, 19, 19, 19, 18, 18, 18, 17, 17, 17, 17, 16, 16, 16 };

static short counter_saw[16];
static short counter_sh[16];
static short counter_tri[16];
static short counter_sine[16];
static short counter_pulse[16];
static short buffer_sh[16];
static short ADSR_Mode[16];
static int ADSR_Value[16];
static int ADSR_SustainCounter[16];
static short (*buffern)[2048] = NULL;
static short filterBuffer[16 * 4];


// VU Bar stuff
#define VUBAR_DECAY_DELAY 1

struct VUBar 
{
	WORD vuValue;
	WORD vuDelay;
};

struct VUBar VUBar1 = {0,0};
struct VUBar VUBar2 = {0,0};
struct VUBar VUBar3 = {0,0};
struct VUBar VUBar4 = {0,0};


UWORD vubar1_colors[16] = { BACKGROUND_COLOR, 0x000, 0x100, 0x200, 0x211, 0x311, 0x322, 0x422, 0x533, 0x644, 0x755, 0x866, 0x977, 0xb99, 0xdbb, 0xfdd };
UWORD vubar2_colors[16] = { BACKGROUND_COLOR, 0x000, 0x110, 0x220, 0x221, 0x331, 0x332, 0x442, 0x553, 0x664, 0x775, 0x886, 0x997, 0xbb9, 0xddb, 0xffd };
UWORD vubar3_colors[16] = { BACKGROUND_COLOR, 0x000, 0x010, 0x020, 0x121, 0x131, 0x232, 0x242, 0x353, 0x464, 0x575, 0x686, 0x797, 0x9b9, 0xbdb, 0xdfd };
UWORD vubar4_colors[16] = { BACKGROUND_COLOR, 0x000, 0x001, 0x002, 0x112, 0x113, 0x223, 0x224, 0x335, 0x446, 0x557, 0x668, 0x779, 0x99b, 0xbbd, 0xddf }; 
//UWORD vubar1_colors[16] = { BACKGROUND_COLOR, 0x000, 0x001, 0x002, 0x112, 0x113, 0x223, 0x224, 0x335, 0x446, 0x557, 0x668, 0x779, 0x99b, 0xbbd, 0xfff };
//UWORD vubar2_colors[16] = { BACKGROUND_COLOR, 0x000, 0x001, 0x002, 0x112, 0x113, 0x223, 0x224, 0x335, 0x446, 0x557, 0x668, 0x779, 0x99b, 0xbbd, 0xfff };
//UWORD vubar3_colors[16] = { BACKGROUND_COLOR, 0x000, 0x001, 0x002, 0x112, 0x113, 0x223, 0x224, 0x335, 0x446, 0x557, 0x668, 0x779, 0x99b, 0xbbd, 0xfff };
//UWORD vubar4_colors[16] = { BACKGROUND_COLOR, 0x000, 0x001, 0x002, 0x112, 0x113, 0x223, 0x224, 0x335, 0x446, 0x557, 0x668, 0x779, 0x99b, 0xbbd, 0xfff };



// Forward references
void GenerateSamples();


//////////////////////////////////////////////////////////////////////////////

/* Get VBR on non 68000 systems (music replayer needs it) */
static APTR GetVBR(void) {
	APTR vbr = 0;
	UWORD getvbr[] = { 0x4e7a, 0x0801, 0x4e73 }; // MOVEC.L VBR,D0 RTE

	if (SysBase->AttnFlags & (AFF_68010|AFF_68020|AFF_68030|AFF_68040|AFF_68060)) 
		vbr = (APTR)Supervisor((ULONG (*)())getvbr);

	return vbr;
}


void SetInterruptHandler(APTR interrupt) {
	*(volatile APTR*)(((UBYTE*)VBR)+0x6c) = interrupt;
}

APTR GetInterruptHandler() {
	return *(volatile APTR*)(((UBYTE*)VBR)+0x6c);
}


//vblank begins at vpos 312 hpos 1 and ends at vpos 25 hpos 1
//vsync begins at line 2 hpos 132 and ends at vpos 5 hpos 18 
void WaitVbl() {
	while (1) {
		volatile ULONG vpos=*(volatile ULONG*)0xDFF004;
		vpos&=0x1ff00;
		if (vpos!=(311<<8))
			break;
	}
	while (1) {
		volatile ULONG vpos=*(volatile ULONG*)0xDFF004;
		vpos&=0x1ff00;
		if (vpos==(311<<8))
			break;
	}
}


inline void WaitBlt() {
	UWORD tst=*(volatile UWORD*)&custom->dmaconr; //for compatiblity a1000
	(void)tst;
	while (*(volatile UWORD*)&custom->dmaconr&(1<<14)) {} //blitter busy wait
}


inline short MouseLeft(){return !((*(volatile UBYTE*)0xbfe001)&64);}	


void TakeSystem() {
	ActiView=GfxBase->ActiView; //store current view
	OwnBlitter();
	WaitBlt();	
	Disable();
	
	//Save current interrupts and DMA settings so we can restore them upon exit. 
	SystemADKCON=custom->adkconr;
	SystemInts=custom->intenar;
	SystemDMA=custom->dmaconr;
	custom->intena=0x7fff;//disable all interrupts
	custom->intreq=0x7fff;//Clear any interrupts that were pending
	
	WaitVbl();
	WaitVbl();
	custom->dmacon=0x7fff;//Clear all DMA channels

	//set all colors black
	for(int a=0;a<32;a++)
		custom->color[a]=0;

	LoadView(0);
	WaitTOF();
	WaitTOF();

	WaitVbl();
	WaitVbl();

	VBR=GetVBR();
	SystemIrq=GetInterruptHandler(); //store interrupt register
}


void FreeSystem() { 
	WaitVbl();
	WaitBlt();
	custom->intena=0x7fff;//disable all interrupts
	custom->intreq=0x7fff;//Clear any interrupts that were pending
	custom->dmacon=0x7fff;//Clear all DMA channels

	//restore interrupts
	SetInterruptHandler(SystemIrq);

	/*Restore system copper list(s). */
	custom->cop1lc=(ULONG)GfxBase->copinit;
	custom->cop2lc=(ULONG)GfxBase->LOFlist;
	custom->copjmp1=0x7fff; //start coppper

	/*Restore all interrupts and DMA settings. */
	custom->intena=SystemInts|0x8000;
	custom->dmacon=SystemDMA|0x8000;
	custom->adkcon=SystemADKCON|0x8000;

	LoadView(ActiView);
	WaitTOF();
	WaitTOF();
	WaitBlt();	
	DisownBlitter();
	Enable();
}

// GCC Assembler syntax
// https://gcc.gnu.org/onlinedocs/gcc/Extended-Asm.html

// Install cia interrupt handler
void mt_install_cia(void)
{
	/* mt_install_cia(a6=CUSTOM, a0=VectorBase, d0=PALflag.b) */
	{
		register int _d0 asm("d0") = 1;
		register const void* _a0 asm("a0") = VBR;
		register const void* _a6 asm("a6") = (void*)0xdff000;

		asm volatile (
			"movem.l %%d0-%%d7/%%a0-%%a6,-(%%sp)\n"
			"jsr _mt_install_cia\n"
			"movem.l (%%sp)+,%%d0-%%d7/%%a0-%%a6"
			: 					// no outputs
			: "r" (_d0), "r" (_a0), "r" (_a6)	// inputs
			: "cc", "memory"			// changes memory and condition codes (can also put regs trashed here "d0", "d1")
			);
	}
}


// Remove cia interrupt handler
void mt_remove_cia(void)
{
	/* mt_remove_cia(a6=CUSTOM) */
	{
		register const void* _a6 asm("a6") = (void*)0xdff000;

		asm volatile (
			"movem.l %%d0-%%d7/%%a0-%%a6,-(%%sp)\n"
			"jsr _mt_remove_cia\n"
			"movem.l (%%sp)+,%%d0-%%d7/%%a0-%%a6"
			: 					// no outputs
			: "r" (_a6)				// inputs
			: "cc", "memory"			// changes memory and condition codes (can also put regs trashed here "d0", "d1")
			);
	}
}


// Init module and start playing immediately
void mt_init(void* ModAdr, void* SmpAdr)
{
	/* _mt_init(a6=CUSTOM, a0=TrackerModule, a1=Samples|NULL, d0=InitialSongPos.b) */
	{
		register int _d0 __asm("d0") = 0;
		register const void* _a0 asm("a0") = ModAdr;
		register const void* _a1 asm("a1") = (void*)SmpAdr;
		register const void* _a6 asm("a6") = (void*)0xdff000;

		asm volatile (
			"movem.l %%d0-%%d7/%%a0-%%a6,-(%%sp)\n"
			"jsr _mt_init\n"
			"move.b #1,_mt_Enable\n"
			"movem.l (%%sp)+,%%d0-%%d7/%%a0-%%a6"			
			: 					// no outputs
			: "r" (_d0), "r" (_a0), "r" (_a1), "r" (_a6)	// inputs
			: "cc", "memory"			// changes memory and condition codes (can also put regs trashed here "d0", "d1")
			);
	}
}

// End module and cleanup
void mt_end(void)
{
	/* mt_end(a6=CUSTOM) */
	{
		register const void* _a6 asm("a6") = (void*)0xdff000;

		asm volatile (
			"movem.l %%d0-%%d7/%%a0-%%a6,-(%%sp)\n"		
			"jsr _mt_end\n"
			"movem.l (%%sp)+,%%d0-%%d7/%%a0-%%a6"				
			: 					// no outputs
			: "r" (_a6)				// inputs
			: "cc", "memory"			// changes memory and condition codes
			);
	}
}

// Init module and start playing immediately
UBYTE mt_getvumeter(void)
{
	UBYTE	mask = 0;

	{
		asm volatile (
			"move.b _mt_VUMeter,%0\n"
			"clr.b _mt_VUMeter"
			: "=g"(mask)				// outputs, g is any type of memory or register, "=" is "a variable overwriting an existing value"
			: 					// no inputs
			: "cc"					// condition codes trashed
			);
	}

	return mask;
}


// Copperlist to use when playing - simple horizontal colors
//bar1 80
//bar2 98
//bar3 b0
//bar4 c8
UWORD copperlist_play[] __attribute__((section (".MEMF_CHIP"))) = {
	0x8007, 0xfffe,			// 0
	0x180, 0x0333,			// 2
	0x9007, 0xfffe,			// 4
	0x180, BACKGROUND_COLOR,	// 6

	0x9807, 0xfffe,			// 8
	0x180, 0x0333,			// 10
	0xa807, 0xfffe,			// 12
	0x180, BACKGROUND_COLOR,	// 14

	0xb007, 0xfffe,			// 16
	0x180, 0x0333,			// 18
	0xc007, 0xfffe,			// 20
	0x180, BACKGROUND_COLOR,	// 22

	0xc807, 0xfffe,			// 24
	0x180, 0x0333,			// 26
	0xd807, 0xfffe,			// 28
	0x180, BACKGROUND_COLOR,	// 30

	0xffff, 0xfffe // end copper list
};

UWORD GetBarColor(struct VUBar *bar, UWORD *barcolors)
{
	// Decrease value if needed
	if (bar->vuValue)
	{
		if (bar->vuDelay <= 0)
		{
			bar->vuValue--;
			bar->vuDelay = VUBAR_DECAY_DELAY;
		}
		else
		{
			bar->vuDelay--;
		}
	}
	
	// For the given bar value, get the color
	return barcolors[bar->vuValue];
}

// IRQ Handler for playing music
static __attribute__((interrupt)) void irqPlay() 
{
	// In our player copperlist the 4 bar colors are stored in words 3, 11, 19, 27

	UBYTE mask = mt_getvumeter();

	if (mask&1)
	{
		VUBar1.vuValue = 15;
		VUBar1.vuDelay = VUBAR_DECAY_DELAY;
	}
	if (mask&2)
	{
		VUBar2.vuValue = 15;
		VUBar2.vuDelay = VUBAR_DECAY_DELAY;
	}
	if (mask&4)
	{
		VUBar3.vuValue = 15;
		VUBar3.vuDelay = VUBAR_DECAY_DELAY;
	}
	if (mask&8)
	{
		VUBar4.vuValue = 15;
		VUBar4.vuDelay = VUBAR_DECAY_DELAY;
	}

	copperlist_play[3] = GetBarColor(&VUBar1, vubar1_colors);
	copperlist_play[11] = GetBarColor(&VUBar2, vubar2_colors);
	copperlist_play[19] = GetBarColor(&VUBar3, vubar3_colors);
	copperlist_play[27] = GetBarColor(&VUBar4, vubar4_colors);

	// Check for early quit if still precalcing
	if (MouseLeft())
		ForceQuit = 1;

	custom->intreq=(1<<INTB_VERTB); 
	custom->intreq=(1<<INTB_VERTB); //reset vbl req. twice for a4000 bug.
}


// Copperlist to use while precalc occurs
UWORD copperlist_precalc[] __attribute__((section (".MEMF_CHIP"))) = {
	0xab4f, 0xfffe, 						//Wait for line ab, entry 0
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	//entry 2
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, BACKGROUND_COLOR,	

	0xac4f, 0xfffe, 						//Wait for line ac, entry 66
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	//entry 68
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	0x180, 0x0333,	
	0x180, BACKGROUND_COLOR,	

	0xffff, 0xfffe // end copper list
};

// IRQ Handler for precalc 
static __attribute__((interrupt)) void irqPrecalc() 
{
	UWORD	entry = 3;	//word 3 is the first color
	static WORD swapToPlayerCountDown = 2;

	for (UWORD i=0; i<=InstrumentsCurrent && i<=31; ++i)
	{
		copperlist_precalc[entry] = 0xccc;
		copperlist_precalc[entry+68] = 0xccc;	//Next line
		entry += 2;	//next color
	}

	// Swap to full bar if ready to move onto playing
	if (InstrumentsCurrent >= EARLYPLAY_INSTRUMENTS_COMPLETE)
	{
		--swapToPlayerCountDown;
		
		entry = 3;	//word 3 is the first color
		for (UWORD i=0; i<=31; ++i)
		{
			copperlist_precalc[entry] = 0xfff;
			copperlist_precalc[entry+68] = 0xfff;	//Next line
			entry += 2;	//next color
		}
	}

	if (swapToPlayerCountDown < 0 && ForceQuit == 0)
	{
		// If enough instruments have precalced switch to player routine and init module
		// Swap to player copperlist and irq
		SetInterruptHandler((APTR)irqPlay);
		custom->cop1lc = (ULONG)copperlist_play;

		// Install cia handler and init module and start playing 
		mt_install_cia();
		mt_init(ModAdr, 0);	// 0 assumes samples at end of module, otherwise can point somewhere else
	}

	// Check for early quit
	if (MouseLeft())
		ForceQuit = 1;

	custom->intreq=(1<<INTB_VERTB); 
	custom->intreq=(1<<INTB_VERTB); //reset vbl req. twice for a4000 bug.
}


// Allocate memory for module plus samples in chip mem and generate the samples
void GenerateSamples(void)
{
	ModAdr = AllocMem((gen_length+mod_length_empty) * sizeof(UBYTE), MEMF_CHIP);
	ImpAdr = AllocMem(imp_length*sizeof(UBYTE), MEMF_CHIP);
	buffern= AllocMem(24*2048*sizeof(short),MEMF_CLEAR);

	// Copy the module and imported samples into chip ram
	memcpy(ModAdr,protrackermod,mod_length_empty);
	memcpy(ImpAdr,importedsamples, imp_length);

	#include "synthnodes.h"

	int SmpLength[31];	
	int ImpLength[8];
	WORD repeat_offset[31];							
	WORD repeat_length[31];								
	UBYTE samplename_flag[31];							

	#include "Ilen.h"

	int inst_count = numinstruments ;				
	int instrument;										
	int smp;										
	short v1,v2,v3,v4;								

	void* BaseAdr[32];
	BaseAdr[0] = ModAdr + mod_length_empty;									
	for (instrument=0;instrument<inst_count;instrument++)					
	{
		BaseAdr[instrument+1]=  BaseAdr[instrument]+ SmpLength[instrument];	
	}

	BYTE last = 0;
	for (int i = 0; i < imp_length; i++)
	{
		BYTE delta = *(BYTE*)(ImpAdr + i);
		*(BYTE*)(ImpAdr + i) = delta + last;
		last = *(BYTE*)(ImpAdr + i);
	}

	void* BaseImpAdr[8];
	BaseImpAdr[0] = ImpAdr;
	for (int k = 0; k<7; k++)
	{
		BaseImpAdr[k + 1] = BaseImpAdr[k] + ImpLength[k];
	}


	for (instrument=0; instrument<inst_count && ForceQuit==0;instrument++)				
	{
		clr_buf();
		v1=0;v2=0;v3=0;v4=0;								
		for (smp=0;smp<=SmpLength[instrument];smp++)		
		{
			#include "Inst.h"
			v1>>=8;
			*(BYTE*)(BaseAdr[instrument]+smp) = v1;		
		}

		if (samplename_flag[instrument]=='l')
		{
			loopgen (repeat_length[instrument],repeat_offset[instrument],BaseAdr[instrument]);
		}
		*(BYTE*)(BaseAdr[instrument]) = 0;		
		*(BYTE*)(BaseAdr[instrument]+1) = 0;		

		InstrumentsCurrent++;		// Update progress
	}

	// Set InstrumentsCurrent to 32 to indicate processed intruments 0-31 for the vbl interrupts
	InstrumentsCurrent = 32;

	// No longer need this buffer.
	FreeMem(buffern, 24*2048*sizeof(short));

}


/*
 * MAIN ENTRY POINT
 */

int main() 
{
	SysBase = *((struct ExecBase**)4UL);
	custom = (struct Custom*)0xdff000;

	// We will use the graphics library only to locate and restore the system copper list once we are through.
	GfxBase = (struct GfxBase *)OpenLibrary((CONST_STRPTR)"graphics.library",0);
	if (!GfxBase)
		Exit(0);

	DOSBase = (struct DosLibrary*)OpenLibrary("dos.library", 0);
	if (!DOSBase) Exit(0);


	// Setup mod stuff
	Write(Output(), "\n",1);		
	Write(Output(), "(c) by Virgill 2022\n", 20);
	Write(Output(), "\n", 1);
	Write(Output(), "AMIGAKLANG v1.00\n",17);		
	Write(Output(), "\n",1);	

	//Write(Output(), "Press lmb to exit when playing.\n",32);		
	Delay(150);

	// System down
	TakeSystem();
	WaitVbl();

	// Enable dma and interrupts
	SetInterruptHandler((APTR)irqPrecalc);
	custom->dmacon = DMAF_SETCLR | DMAF_MASTER | DMAF_COPPER;
	custom->intena=(1<<INTB_SETCLR)|(1<<INTB_INTEN)|(1<<INTB_VERTB);
	custom->cop1lc = (ULONG)copperlist_precalc;

	// Generate the samples
	GenerateSamples();

	// Wait for mouse click to quit if allowed to generate all samples (user didn't pre-quit during precalc)
	while(ForceQuit == 0 && !MouseLeft()) 
	{
	}

	// End music and remove cia handler
	mt_end();
	mt_remove_cia();

	// Restore system
	FreeSystem();

	// Free memory
	FreeMem(ImpAdr, imp_length*sizeof(UBYTE));
	FreeMem(ModAdr, (gen_length+mod_length_empty) * sizeof(UBYTE));

	CloseLibrary((struct Library*)DOSBase);
	CloseLibrary((struct Library*)GfxBase);

	Exit(0);
}


//////////////////////
#endif //executable
